﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace VisualGitSafe
{
    public static class Extensions
    {
        public static User Find(this List<User> that,string userName)
        {
            foreach(User user in that)
            {
                if (String.Compare(user.UserName,userName, true) == 0)
                {
                    return user;
                }
            }
            return null;
        }

        /// <summary>
        /// Sets the IsReadOnly flag on every item in the list.
        /// </summary>
        /// <param name="that"></param>
        /// <param name="isReadOnly"></param>
        public static void SetIsReadOnlyFlag(this List<ReadOnlyFile> that,bool isReadOnly)
        {
            foreach(ReadOnlyFile loop in that)
            {
                loop.IsReadOnly = isReadOnly;
            }
        }

        /// <summary>
        /// Adds a file to the list of read only files. Makes the file read only.
        /// </summary>
        /// <param name="that"></param>
        /// <param name="fileToAdd">File to add</param>
        /// <param name="programRootFolder">Program Root Folder from configuration</param>
        public static void Add(this List<ReadOnlyFile> that,string fileToAdd, string programRootFolder)
        {
            ReadOnlyFile newFile;
            bool found;
            found = false;
            foreach (ReadOnlyFile file in that)
            {
                if (String.Compare(file.FileName, fileToAdd, true) == 0)
                {
                    // This file is already in the list of read only files.
                    file.IsReadOnly = true;
                    found = true;
                    break;
                }
            }
            if (!found)
            {
                // This is a new conflict.
                // Add this file to the list of read only files and set the file to read only.
                newFile = new ReadOnlyFile();
                newFile.FileName = fileToAdd;
                newFile.SetReadOnly(programRootFolder, true);
                that.Add(newFile);
                newFile = null;
            }
        }

        /// <summary>
        /// Returns a list of items in a good format for ListBox.Items.AddRange()
        /// </summary>
        /// <param name="that"></param>
        /// <returns>Returns the file names in a format to display in a list box</returns>
        public static object[] Items(this List<MyFile> that)
        {
            object[] items = null;

            if ((that != null) && (that.Count > 0))
            {
                items = (from item in that
                         select item.FileName).ToArray();
            }
            return items;
        }

        /// <summary>
        /// Returns a list of items in a good format for ListBox.Items.AddRange()
        /// </summary>
        /// <param name="that"></param>
        /// <returns>Returns the file names in a format to display in a list box</returns>
        public static object[] Items(this List<OtherPeoplesFile> that)
        {
            object[] items = null;

            if ((that != null) && (that.Count > 0))
            {
                items = (from item in that
                         select $"{item.FileName} {item.Users}" ).ToArray();
            }
            return items;
        }

        /// <summary>
        /// Returns the index of fileName in the list of files.
        /// </summary>
        /// <param name="that"></param>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public static int IndexOf(this List<OtherPeoplesFile> that,string fileName)
        {
            int index;
            if ((that != null) && (that.Count > 0))
            {
                for(index=0;index<that.Count;index++)
                {
                    if (String.Compare(that[index].FileName,fileName,true)==0)
                    {
                        return index;
                    }
                }
            }

            return -1;
        }

        /// <summary>
        /// Returns the index of fileName in the list of files.
        /// </summary>
        /// <param name="that"></param>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public static int IndexOf(this List<MyFile> that, string fileName)
        {
            int index;
            if ((that != null) && (that.Count > 0))
            {
                for (index = 0; index < that.Count; index++)
                {
                    if (String.Compare(that[index].FileName, fileName, true) == 0)
                    {
                        return index;
                    }
                }
            }

            return -1;
        }
    }
}
