﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;
using System.Threading;
using System.Diagnostics;

namespace VisualGitSafe
{
    public partial class Form1 : Form
    {
        private Configuration _configuration;
        private List<FileSystemWatcher> _watchers;
        private List<MyFile> _myFiles;
        private List<OtherPeoplesFile> _otherPeoplesFiles;
        private List<User> _users;
        private Thread _updateFilesThread;
        private List<ReadOnlyFile> _readOnlyFiles;

        private const string HOOK_FOLDER = ".git\\hooks";
        //private Thread _doubleCheckThread;
        private object _directoryScan;
        private const string VERSION = "2.1";

        /// <summary>
        /// Set this indicator to cause the background thread to run git status to get the list of changed files.
        /// </summary>
        private System.Threading.ManualResetEventSlim _fileChangedIndicator;

        /// <summary>
        /// Set this to true to cause the update files background thread to close.
        /// </summary>
        private volatile bool _closeUpdateFilesThread;

        /// <summary>
        /// Resize controls on the form when the form resizes.
        /// </summary>
        private AspectRatio _aspectRatio;
        
        public Form1()
        {
            string gitFile;
            bool foundConfigurationFile;
            _myFiles = new List<MyFile>();
            _otherPeoplesFiles = new List<OtherPeoplesFile>();
            _directoryScan = new object();
            InitializeComponent();
            foundConfigurationFile = LoadConfiguration();          
            _readOnlyFiles = new List<ReadOnlyFile>();
            if ((!String.IsNullOrEmpty(_configuration.SharedFolder)) 
                && (!String.IsNullOrEmpty(_configuration.YourName)) 
                &&  (Directory.Exists(_configuration.SharedFolder)))
            {
                // Delete the git file.
                gitFile = Path.Combine(_configuration.SharedFolder, _configuration.YourName) + ".git";
                if (File.Exists(gitFile))
                {
                    File.Delete(gitFile);
                }
            }
            _updateFilesThread = null;
            _fileChangedIndicator = new ManualResetEventSlim(false);
            _watchers = new List<FileSystemWatcher>();
            _users = new List<User>();
            _closeUpdateFilesThread = false;

            ReadAllUserChanges();
            MarkOtherPeoplesChangedFilesReadOnly();
            DisplayOtherPeoplesChangedFiles();
            StartWatching();
            StartFileChangedThread(); // Update the list of changed files.
            //StartDoubleCheckThread();
            _aspectRatio = new AspectRatio(this);
            _aspectRatio.AddControl(grpMyChangedFiles, 0, 1, 0, 0.5);
            _aspectRatio.AddControl(lstMyChangedFiles, 0, 1, 0, 0.5);
            _aspectRatio.AddControl(grpOtherChangedFiles, 0, 1, 0.5, 0.5);
            _aspectRatio.AddControl(lstOtherChangedFiles, 0, 1, 0, 0.5);
            if (!foundConfigurationFile)
            {
                if (MessageBox.Show($"Welcome to Visual Git Safe {VERSION}.\nYou need to enter configuration settings first.\nWould you like to do that now?",$"Visual Git Safe {VERSION}",MessageBoxButtons.YesNo)==DialogResult.Yes)
                {
                    mnuConfigure_Click(null, null);
                }
            }
        }

        /*private void StartDoubleCheckThread()
        {
            ThreadStart start = new ThreadStart(DoubleCheck);
            _doubleCheckThread = new Thread(start);
            _doubleCheckThread.IsBackground = true;
            _doubleCheckThread.Start();
        }*/

        /// <summary>
        /// Runs in a separate thread. Double check the shared files occasionally.
        /// </summary>
        /*private void DoubleCheck()
        {
            const int THIRTY_SECONDS = 3000;
            int loop;
            DirectoryInfo sharedFolder;
            FileInfo[] files;
            string extension;
            string fileName;
            string fileNameWithoutExtension;
            User user;
            bool updatedUsers;

            while (!_closeUpdateFilesThread)
            {
                // Wait 30 seconds.
                try
                {
                    lock (_directoryScan)
                    {
                        updatedUsers = false;
                        if ((!String.IsNullOrEmpty(_configuration.SharedFolder)) && (Directory.Exists(_configuration.SharedFolder)))
                        {                            
                            sharedFolder = new DirectoryInfo(_configuration.SharedFolder);
                            files = sharedFolder.GetFiles();
                            foreach (FileInfo file in files)
                            {
                                if (_closeUpdateFilesThread)
                                {
                                    break;
                                }
                                fileName = file.Name;
                                extension = Path.GetExtension(fileName);
                                if (!String.IsNullOrEmpty(_configuration.YourName))
                                {
                                    if (string.Compare(extension, ".git", true) == 0)
                                    {
                                        fileNameWithoutExtension = Path.GetFileNameWithoutExtension(fileName);
                                        if (String.Compare(fileNameWithoutExtension, _configuration.YourName, true) == 0)
                                        {
                                            file.Delete();
                                            StartFileChangedThread();
                                        }
                                    }
                                    continue;
                                }                                
                                if (String.Compare(extension,".txt",true) == 0)
                                {
                                    fileNameWithoutExtension = Path.GetFileNameWithoutExtension(fileName);
                                    lock(_users)
                                    {
                                        user = _users.Find(fileNameWithoutExtension);
                                    }
                                    if (user == null)
                                    {
                                        updatedUsers = true;
                                        user = new User();
                                        user.UserName = fileNameWithoutExtension;
                                        user.LastModified = file.LastWriteTime;
                                        user.ChangedFiles = File.ReadAllLines(file.FullName).ToList();
                                        lock(_users)
                                        {
                                            _users.Add(user);
                                        }
                                    } else
                                    {
                                        if (user.LastModified != file.LastWriteTime)
                                        {
                                            updatedUsers = true;
                                            user.LastModified = file.LastWriteTime;
                                            user.ChangedFiles = File.ReadAllLines(file.FullName).ToList();
                                        }
                                    }
                                }
                            }
                        }
                        if ((updatedUsers) && (!_closeUpdateFilesThread))
                        {
                            ShowAllUserChanges();
                        }
                    }
                    // Wait 30 seconds.
                    for (loop = 0; loop < THIRTY_SECONDS; loop++)
                    {
                        Thread.Sleep(10);
                        if (_closeUpdateFilesThread)
                        {
                            break;
                        }
                    }
                } catch(Exception)
                {

                }
            }
            _doubleCheckThread = null;
        }*/

        private void ReadAllUserChanges()
        {
            DirectoryInfo sharedFolder;
            FileInfo[] files;
            try
            {
                if ((!String.IsNullOrEmpty(_configuration.SharedFolder)) && (Directory.Exists(_configuration.SharedFolder)))
                {
                    sharedFolder = new DirectoryInfo(_configuration.SharedFolder);
                    files = sharedFolder.GetFiles("*.txt");
                    foreach(FileInfo loop in files)
                    {
                        ReadUserChange(loop.FullName, loop.LastWriteTime);
                    }
                }
            }
            catch (Exception)
            {

            }
        }

        private void MarkOtherPeoplesChangedFilesReadOnly()
        {
            int count;
            int index;
            try
            {
                if ((_configuration.MakeEditedFilesReadOnly) 
                    && (!String.IsNullOrEmpty(_configuration.ProgramRootFolder)) 
                    && (_users != null))
                {
                    lock (_readOnlyFiles)
                    {
                        _readOnlyFiles.SetIsReadOnlyFlag(false);
                        lock (_users)
                        {
                            foreach (User user in _users)
                            {
                                if (String.Compare(user.UserName, _configuration.YourName, true) == 0)
                                {
                                    continue;
                                }
                                if ((user.ChangedFiles != null) && (user.ChangedFiles.Count > 0))
                                {
                                    foreach (string changedFile in user.ChangedFiles)
                                    {
                                        _readOnlyFiles.Add(changedFile, _configuration.ProgramRootFolder);
                                    }                                    
                                }
                            }
                        }
                        // Turn off any files that are no longer read only.
                        count = _readOnlyFiles.Count;
                        for (index = count - 1; index >= 0; index--)
                        {
                            if (!_readOnlyFiles[index].IsReadOnly)
                            {
                                _readOnlyFiles[index].SetReadOnly(_configuration.ProgramRootFolder);
                                _readOnlyFiles.RemoveAt(index);
                            }
                        }
                    }
                }
            } catch(Exception)
            {

            }
        }

        /// <summary>
        /// Show the files from other users.
        /// </summary>
        private void DisplayOtherPeoplesChangedFiles()
        {
            List<OtherPeoplesFile> output;
            object[] items;
            int index;
            int myIndex;
            bool conflict;

            if (InvokeRequired)
            {
                BeginInvoke((Action)DisplayOtherPeoplesChangedFiles);
            } else
            {                
                output = new List<OtherPeoplesFile>();
                lock (_myFiles)
                {
                    foreach(MyFile myfile in _myFiles)
                    {
                        myfile.HasConflict = false;
                    }
                }
                lock (_users)
                {
                    foreach (User user in _users)
                    {
                        if ((String.IsNullOrEmpty(_configuration.YourName)) || (String.Compare(user.UserName, _configuration.YourName, true) != 0))
                        {
                            foreach (string fileName in user.ChangedFiles)
                            {
                                index = output.IndexOf(fileName);
                                if (index >= 0)
                                {
                                    output[index].Users = $"{output[index].Users} {user.UserName}";
                                    output[index].MinorConflict = true;
                                }
                                else
                                {
                                    // Figure out if there's a conflict.
                                    lock (_myFiles)
                                    {
                                        myIndex = _myFiles.IndexOf(fileName);
                                        if (myIndex >=0)
                                        {
                                            _myFiles[myIndex].HasConflict = true;
                                        }
                                    }
                                    conflict = myIndex >= 0;
                                    // Add this file name to the list of files.
                                    output.Add(new OtherPeoplesFile(fileName, user.UserName, conflict,false));                                    
                                }
                            }
                        }
                    }
                }
                // Sort the list of files by file name.
                if (output.Count >= 2)
                {
                    output = (from file in output
                              orderby file.FileName
                              select file
                              ).ToList();
                }
                          
                // display the list of files.
                items = output.Items();
                lock(_otherPeoplesFiles)
                {
                    _otherPeoplesFiles.Clear();
                    _otherPeoplesFiles.AddRange(output);
                }
                lstOtherChangedFiles.Items.Clear();
                if (items != null)
                {
                    lstOtherChangedFiles.Items.AddRange(items);
                }
                lstMyChangedFiles.Refresh();
            }
        }

        private void ReadUserChange(string fullPath, DateTime lastModified)
        {
            string userName = Path.GetFileNameWithoutExtension(fullPath);
            List<string> filesChanged;
            if (File.Exists(fullPath))
            {
                filesChanged = File.ReadAllLines(fullPath).ToList();
            } else
            {
                // The file was deleted.
                filesChanged = new List<string>();
            }
            ReadUserChange(userName, filesChanged, lastModified);
        }

        private void ReadUserChange(string userName, List<string> filesChanged,DateTime lastModified)
        {
            User newUser;
            lock (_users)
            {
                foreach (User loop in _users)
                {
                    if (String.Compare(loop.UserName, userName, true) == 0)
                    {
                        loop.ChangedFiles = filesChanged;
                        loop.LastModified = lastModified;
                        return;
                    }
                }
                newUser = new User();
                newUser.UserName = userName;
                newUser.LastModified = lastModified;
                newUser.ChangedFiles = filesChanged;
                _users.Add(newUser);
            }
        }

        private void ReadUserChange(string fullPath, User user)
        {
            user.ChangedFiles = File.ReadAllLines(fullPath).ToList();
        }

        private void StartWatching()
        {
            DirectoryInfo programRootFolder;
            DirectoryInfo sharedRootFolder;
            if (String.IsNullOrEmpty(_configuration.ProgramRootFolder))
            {
                return;
            }
            if (!Directory.Exists(_configuration.ProgramRootFolder))
            {
                return;
            }
            programRootFolder = new DirectoryInfo(_configuration.ProgramRootFolder);
            AddFileSystemWatchers(programRootFolder, true);
            if ((!String.IsNullOrEmpty(_configuration.SharedFolder)) 
                && (Directory.Exists(_configuration.SharedFolder)))
            {
                sharedRootFolder = new DirectoryInfo(_configuration.SharedFolder);
                AddFileSystemWatchers(sharedRootFolder, false);
            }
        }

        private void AddFileSystemWatchers(DirectoryInfo folderToWatch, bool recurseFolders)
        {
            try
            {
                FileSystemWatcher watcher;
                //DirectoryInfo[] folders;
                watcher = new FileSystemWatcher(folderToWatch.FullName);
                watcher.NotifyFilter = NotifyFilters.LastWrite | NotifyFilters.CreationTime | NotifyFilters.Size | NotifyFilters.FileName | NotifyFilters.Attributes;
                watcher.Filter = "*.*";
                watcher.Changed += Watcher_Changed;
                watcher.Created += Watcher_Created;
                watcher.Deleted += Watcher_Deleted;
                watcher.EnableRaisingEvents = true;
                watcher.IncludeSubdirectories = recurseFolders;

                _watchers.Add(watcher);
                watcher = null;
            } catch(Exception)
            {
                
            }
        }

        private void Watcher_Deleted(object sender, FileSystemEventArgs e)
        {
            FileChanged(e.FullPath);
        }

        private void Watcher_Created(object sender, FileSystemEventArgs e)
        {
            FileChanged(e.FullPath);
        }

        private void mnuConfigure_Click(object sender, EventArgs e)
        {
            frmConfiguration frmConfiguration = new frmConfiguration(_configuration);
            StopWatching();
            ClearReadOnlyFiles();
            frmConfiguration.ShowDialog(this);
            if (_configuration.IsDirty)
            {                
                SaveConfiguration();                
            }
            StartWatching();
            StartFileChangedThread();
            MarkOtherPeoplesChangedFilesReadOnly();            
            CheckForGitHooks();
        }

        /// <summary>
        /// Check for git hooks to see that they are all there. Asks if they should be added.
        /// </summary>
        private bool CheckForGitHooks()
        {
            bool foundAllHooks = false;
            string hooksFolder;
            string missingHooks;
            KeyValuePair<string, bool>[] hooksFound;
            try
            {
                hooksFolder = Path.Combine(_configuration.ProgramRootFolder, HOOK_FOLDER);
                if (Directory.Exists(hooksFolder))
                {
                    hooksFound = GetExistingHooks(hooksFolder);
                    // Are there any missing hooks.
                    if (hooksFound.Any(x=>!x.Value))
                    {
                        missingHooks = String.Join(",", (from hook in hooksFound
                                                         where !hook.Value
                                                         select hook.Key).ToArray());
                        if (MessageBox.Show($"You are missing the following git hooks:{missingHooks}. Do you want to add them?","Add Git Hooks", MessageBoxButtons.YesNo)==DialogResult.Yes)
                        {
                            AddGitHooks(hooksFound);
                        } else
                        {
                            MessageBox.Show("You can add them by clicking on Configuration/Git Hooks. You can find out more by clicking on Help/Git Hooks");
                        }
                    } else
                    {
                        foundAllHooks = true;
                    }
                }
            }
            catch (Exception)
            {

            }
            return foundAllHooks;
        }

        private void AddGitHooks(KeyValuePair<string,bool>[] hooksFound)
        {
            string hookFile;
            string[] hookData;
            foreach(KeyValuePair<string,bool> hook in hooksFound)
            {
                if (!hook.Value)
                {
                    hookFile = Path.Combine(_configuration.ProgramRootFolder, HOOK_FOLDER, hook.Key);
                    hookData = GetHookData(hook.Key);
                    File.WriteAllLines(hookFile, hookData);
                }
            }
        }

        /// <summary>
        /// Returns the text data that will go in the git hook.
        /// </summary>
        /// <param name="hookName">Name of the git hook file.</param>
        /// <returns>Returns a bash script</returns>
        private string[] GetHookData(string hookName)
        {
            List<string> hookData;
            string description;
            Dictionary<string, string> descriptions = new Dictionary<string, string>(); // When it says Collection initialization can be similified. Don't do it. It'll break the build.

            // Figure out the description.
            descriptions.Add("post-checkout", "git checkout");
            descriptions.Add("post-merge", "git pull");
            descriptions.Add("post-commit", "commit");
            description = (descriptions.ContainsKey(hookName) ? descriptions[hookName] : "git command");

            // Build the hookData.
            hookData = new List<string>(); // When it says Collection initialization can be similified. Don't do it. It'll break the build.
            hookData.Add("#!/bin/sh");
            hookData.Add("# This batch file runs when you do a " + description);
            hookData.Add("INI_FILE=\"" + INIFile().Replace('\\','/') + "\"");
            hookData.Add("if [ -f \"$INI_FILE\" ]; then");
            hookData.Add("  USER=`awk '/Your Name:/{ print substr($0,11) }' $INI_FILE`");
            hookData.Add("  SHARED_FOLDER=`awk '/Shared Folder:/{ print substr($0,15) }' $INI_FILE`");
            hookData.Add("  if [ -d \"$SHARED_FOLDER\" ]; then");
            hookData.Add("    FILE=\"${SHARED_FOLDER}/${USER}.git\"");
            hookData.Add("    if [ ! -f \"$FILE\" ]; then");
            hookData.Add("      echo \"" + hookName + "\" >$FILE");
            hookData.Add("    fi");
            hookData.Add("  fi");
            hookData.Add("fi");
            hookData.Add("exit 0");

            return hookData.ToArray();
        }

        /// <summary>
        /// Returns the list of hooks and whether or not the hook was found.
        /// </summary>
        /// <param name="hooksFolder"></param>
        /// <returns></returns>
        private KeyValuePair<string, bool>[] GetExistingHooks(string hooksFolder)
        {
            string[] hooks;
            DirectoryInfo hooksDirectory;
            FileInfo[] files;
            KeyValuePair<string, bool>[] hooksFound;

            hooks = GetHooksList();
            hooksDirectory = new DirectoryInfo(hooksFolder);
            files = hooksDirectory.GetFiles();
            hooksFound = (from hook in hooks
                          select new KeyValuePair<string, bool>(hook, files.Any(x => x.Name == hook))).ToArray();
            return hooksFound;
        }

        /// <summary>
        /// Returns the list of hook files.
        /// </summary>
        /// <returns>Returns the list of hook file names.</returns>
        private string[] GetHooksList()
        {
            return "post-checkout,post-commit,post-merge".Split(',');
        }

        /// <summary>
        /// Read the configuration data from the .ini file.
        /// <returns>Returns true if a file was found.</returns>
        /// </summary>
        private bool LoadConfiguration()
        {
            bool foundFile;
            string iniFile = INIFile();
            string[] fileData = null;

            _configuration = new Configuration();
            foundFile = File.Exists(iniFile);
            if (foundFile)
            {
                fileData = File.ReadAllLines(iniFile);                
            } else
            {
                _configuration.Extensions = Configuration.DefaultExtensions;
            }
            _configuration.Deserialize(fileData);
            _configuration.IsDirty = false;
            return foundFile;
        }

        private void SaveConfiguration()
        {
            string iniFile = INIFile();
            string[] fileData = _configuration.Serialize();
            File.WriteAllLines(iniFile, fileData);
            _configuration.IsDirty = false;
        }

        private string INIFile()
        {
            // RETURNS AN INI FILE IN THE SAME DIRECTORY AS THE EXE WITH
            // THE SAME NAME AS THE EXE.
            System.Text.StringBuilder sbReturn;

            sbReturn = new System.Text.StringBuilder();
            sbReturn.Append(Application.ExecutablePath);
            sbReturn.Remove(sbReturn.Length - 4, 4);
            sbReturn.Append(".INI");
            return sbReturn.ToString();
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            Hide();
            Application.DoEvents();
            StopWatching();
            CloseUpdateFilesThread();
            ClearReadOnlyFiles();            
        }

        /// <summary>
        /// Turn off the readonly attribute on all the read only files and clear the list.
        /// </summary>
        private void ClearReadOnlyFiles()
        {
            lock (_readOnlyFiles)
            {
                foreach (ReadOnlyFile file in _readOnlyFiles)
                {
                    file.SetReadOnly(_configuration.ProgramRootFolder,false);
                }
                _readOnlyFiles.Clear();
            }
        }

        private void CloseUpdateFilesThread()
        {
            _closeUpdateFilesThread = true;
            _fileChangedIndicator.Set();
        }

        /// <summary>
        /// Runs when somebody changes a file.
        /// </summary>
        /// <param name="source"></param>
        /// <param name="e"></param>
        private void Watcher_Changed(object source, FileSystemEventArgs e)
        {
            // Update the changed file list.
            FileChanged(e.FullPath);
        }

        /// <summary>
        /// This method is run when a file is added, changed, or deleted.
        /// </summary>
        /// <param name="fullFilePath"></param>
        private void FileChanged(string fullFilePath)
        {
            string extension;
            string fileName;
            DateTime lastModified;
            string[] extensions;
            try
            {
                if ((!String.IsNullOrEmpty(_configuration.ProgramRootFolder)) 
                    && (fullFilePath.StartsWith(_configuration.ProgramRootFolder,StringComparison.OrdinalIgnoreCase)))
                {
                    // A program file changed.                    
                    if (!String.IsNullOrEmpty(_configuration.Extensions))
                    {
                        extension = Path.GetExtension(fullFilePath);
                        extensions = _configuration.Extensions.Split(',');
                        if (extensions.Contains(extension))
                        {
                            StartFileChangedThread();
                        }
                    }
                }
                if ((!String.IsNullOrEmpty(_configuration.SharedFolder)) 
                    && (fullFilePath.StartsWith(_configuration.SharedFolder, StringComparison.OrdinalIgnoreCase)))
                {
                    // A file in the shared folder changed.
                    fileName = Path.GetFileName(fullFilePath);
                    extension = Path.GetExtension(fullFilePath);
                    if (String.Compare(fileName, $"{_configuration.YourName}.git", true) == 0)
                    {
                        // The current user ran git checkout or git push.
                        // This causes the list of files to change.
                        if (File.Exists(fullFilePath))
                        {
                            File.Delete(fullFilePath);
                            StartFileChangedThread();
                        }
                    }
                    // Read somebody else's configuration.
                    if ((String.Compare(extension, ".txt", true) == 0) 
                        && (String.Compare(fileName, $"{_configuration.YourName}.txt", true) != 0))
                    {
                        if (File.Exists(fullFilePath))
                        {
                            lastModified = File.GetLastWriteTime(fullFilePath);
                        } else
                        {
                            lastModified = DateTime.Now;
                        }                        
                        ReadUserChange(fullFilePath, lastModified);
                        MarkOtherPeoplesChangedFilesReadOnly();                        
                        DisplayOtherPeoplesChangedFiles();                        
                    }
                }
            } catch(Exception)
            {

            }
        }
                
        /// <summary>
        /// Start the thread that looks for changed files.
        /// </summary>
        private void StartFileChangedThread()
        {
            ThreadStart start;
            if ((_configuration != null) && (!_configuration.IsDirty) && (!String.IsNullOrEmpty(_configuration.ProgramRootFolder)))
            {
                if (_updateFilesThread == null)
                {
                    start = new ThreadStart(UpdateFilesChanged);
                    _updateFilesThread = new Thread(start);
                    _updateFilesThread.Start();
                }
            }
            // If the thread is already running, tell it to look for changed files.
            _fileChangedIndicator.Set();
        }

        /// <summary>
        /// Runs in a separate thread to update the changed files.
        /// </summary>
        private void UpdateFilesChanged()
        {
            List<string> filesChanged;
            RunProgram program;
            int count;
            int index;
            string fileChanged;
            int search;
            try
            {
                while(!_closeUpdateFilesThread)
                {
                    _fileChangedIndicator.Wait(); // Wait until somebody calls _fileChanged.Set().
                    if (!_closeUpdateFilesThread)
                    {
                        lock(_directoryScan)
                        {
                            program = new RunProgram("git", "status --porcelain", _configuration.ProgramRootFolder);
                            filesChanged = program.Output;
                            // Remove the ?? or the M at the start of the file change.
                            count = (filesChanged == null ? 0 : filesChanged.Count);
                            for(index = 0;index < count; index++)
                            {
                                fileChanged = filesChanged[index];
                                search = 0;
                                if (fileChanged.StartsWith(" "))
                                {
                                    search++;
                                }
                                search = fileChanged.IndexOf(" ", search)+1;
                                filesChanged[index] = (search == 0 ? fileChanged : fileChanged.Substring(search));
                            }
                            ReadUserChange(_configuration.YourName, filesChanged, DateTime.Now);
                            SaveFilesChanged(filesChanged);                            
                            _fileChangedIndicator.Reset();
                            DisplayMyChangedFiles(filesChanged);
                        }                        
                    }
                }                
            } catch (Exception)
            {

            }
            _updateFilesThread = null;
        }

        private void DisplayMyChangedFiles(List<string> filesChanged)
        {
            object[] items;
            List<MyFile> output;
            int index;
            if (InvokeRequired)
            {
                BeginInvoke((Action<List<string>>)DisplayMyChangedFiles, filesChanged);
            } else
            {
                // Convert filesChanged into a list of MyFile objects.
                lock (_otherPeoplesFiles)
                {
                    foreach(OtherPeoplesFile file in _otherPeoplesFiles)
                    {
                        file.MajorConflict = false;
                    }
                    output = new List<MyFile>();
                    foreach (string fileName in filesChanged)
                    {
                        index = _otherPeoplesFiles.IndexOf(fileName);
                        output.Add(new MyFile(fileName, index >= 0));
                        if (index >=0)
                        {
                            _otherPeoplesFiles[index].MajorConflict = true;
                        }
                    }
                }
                // Sort the files.
                if (output.Count >= 2)
                {
                    output = (from file in output
                              orderby file.FileName
                              select file).ToList();
                }
                lock(_myFiles)
                {
                    _myFiles.Clear();
                    _myFiles.AddRange(output);
                    items = _myFiles.Items();
                }                
                lstMyChangedFiles.Items.Clear();
                if (items != null)
                {
                    lstMyChangedFiles.Items.AddRange(items);
                }
                lstOtherChangedFiles.Refresh();
            }
        }

        /// <summary>
        /// Save the list of changed files to the shared drive.
        /// </summary>
        /// <param name="filesChanged"></param>
        private void SaveFilesChanged(List<string> filesChanged)
        {
            string[] existingFiles;
            string fileName;
            bool saveChanges;
            int fileCount;
            int fileLoop;

            try
            {
                saveChanges = false;
                fileName = null;
                if ((!String.IsNullOrEmpty(_configuration.SharedFolder)) 
                    &&  (!String.IsNullOrEmpty(_configuration.YourName)) 
                    && (Directory.Exists(_configuration.SharedFolder)))
                {
                    fileName = Path.Combine(_configuration.SharedFolder, $"{_configuration.YourName}.txt");
                    if (File.Exists(fileName))
                    {
                        existingFiles = File.ReadAllLines(fileName);
                        fileCount = existingFiles == null ? 0 : existingFiles.Length;
                        if (filesChanged.Count != fileCount)
                        {
                            saveChanges = true;
                        }
                        else
                        {
                            for (fileLoop = 0; fileLoop < fileCount; fileLoop++)
                            {
                                if (existingFiles[fileLoop] != filesChanged[fileLoop])
                                {
                                    saveChanges = true;
                                    break;
                                }
                            }
                        }
                    }
                    else
                    {
                        saveChanges = true;
                    }
                }
                if (saveChanges)
                {

                    File.WriteAllLines(fileName, filesChanged);
                }
            } catch(Exception)
            {

            }
        }
                
        private void StopWatching()
        {
            foreach(FileSystemWatcher watcher in _watchers)
            {
                watcher.EnableRaisingEvents = false;
                watcher.Changed -= Watcher_Changed;
                watcher.Created -= Watcher_Created;
                watcher.Deleted -= Watcher_Deleted;
                watcher.Dispose();
            }
            _watchers.Clear();
        }

        private void mnuHelpAbout_Click(object sender, EventArgs e)
        {
            MessageBox.Show($"Visual Git Safe {VERSION}\nAdd features from Visual Source Safe to Git like\n * Knowing what files others are editing.\n * Files others are editing are read only on your computer.\n * Listing conflicts where you and someone else is editing the same file at the same time.","About Visual Git Safe");
        }

        private void mnuSetGitHooks_Click(object sender, EventArgs e)
        {
            if (CheckForGitHooks())
            {
                MessageBox.Show("You already have all the git hooks configured.","Git Hooks");
            }
        }

        private void lstMyChangedFiles_DrawItem(object sender, DrawItemEventArgs e)
        {
            // https://stackoverflow.com/questions/2554609/c-sharp-changing-listbox-row-color
            e.DrawBackground();
            Graphics g = e.Graphics;
            int index = e.Index;
            bool hasConflict;
            string text;
            Color background;
            Color forecolor = Color.Black; // e.ForeColor
            
            lock (_myFiles)
            {
                hasConflict = ((index >= 0) && (_myFiles != null) && (index < _myFiles.Count)) ? _myFiles[index].HasConflict : false;
                text = ((index >= 0) && (_myFiles != null) && (index < _myFiles.Count)) ? _myFiles[index].FileName : String.Empty;
            }
            background = hasConflict ? Color.Red : Color.White;
            // draw the background color you want
            // mine is set to olive, change it to whatever you want
            g.FillRectangle(new SolidBrush(background), e.Bounds);


            // draw the text of the list item, not doing this will only show
            // the background color
            // you will need to get the text of item to display
            g.DrawString(text, e.Font, new SolidBrush(forecolor), new PointF(e.Bounds.X, e.Bounds.Y));

            e.DrawFocusRectangle();
        }

        private void lstOtherChangedFiles_DrawItem(object sender, DrawItemEventArgs e)
        {
            // https://stackoverflow.com/questions/2554609/c-sharp-changing-listbox-row-color
            e.DrawBackground();
            Graphics g = e.Graphics;
            int index = e.Index;
            int conflictColor;
            string text;
            Color background;
            Color forecolor = Color.Black; // e.ForeColor

            lock (_otherPeoplesFiles)
            {
                conflictColor = ((index >= 0) && (_otherPeoplesFiles != null) && (index < _otherPeoplesFiles.Count)) ? _otherPeoplesFiles[index].ConflictColor : 0;
                text = ((index >= 0) && (index < lstOtherChangedFiles.Items.Count)) ? lstOtherChangedFiles.Items[index].ToString() : String.Empty;
            }

            // Figure out the color.
            switch (conflictColor)
            {
                case 1:
                    {
                        background = Color.Tomato;
                        break;
                    }
                case 2:
                    {
                        background = Color.Yellow;
                        break;
                    }
                case 3:
                    {
                        background = Color.Purple;
                        break;
                    }
                case 0:
                default:
                    {
                        background = Color.White;
                        break;
                    }
            }
            // draw the background color you want
            // mine is set to olive, change it to whatever you want
            g.FillRectangle(new SolidBrush(background), e.Bounds);

            // draw the text of the list item, not doing this will only show
            // the background color
            // you will need to get the text of item to display
            g.DrawString(text, e.Font, new SolidBrush(forecolor), new PointF(e.Bounds.X, e.Bounds.Y));

            e.DrawFocusRectangle();
        }

        private void mnuHelpGitHooks_Click(object sender, EventArgs e)
        {
            string[] hooks = GetHooksList();
            MessageBox.Show($"Git hooks are bash scripts that run when you do a git commit, git checkout, or get pull. They are the files {String.Join(",",hooks)} in the {_configuration.ProgramRootFolder}\\{HOOK_FOLDER} folder ","Git Hooks");
        }

        private void mnuHelpConflicts_Click(object sender, EventArgs e)
        {
            MessageBox.Show("Conflicts are files that two people are editing at the same time. Conflicts between you and other people show up as a red background. Conflicts between two or more people that are not you have a yellow background. Conflicts between you and more than one other person have a purple background.", "Conflicts");
        }

        private void mnuHelpReadOnly_Click(object sender, EventArgs e)
        {
            MessageBox.Show("When you check \"Make Other People's Edited Files Read Only\" in the configuration, for any file that someone else is editing, the corresponding file on your computer will be set to read only. When they commit in the file or checkout the file, the read only attribute will be removed from the corresponding file on your computer.", "Read Only Files Option");
        }

        private void mnuOpenGitHooks_Click(object sender, EventArgs e)
        {
            KeyValuePair<string, bool>[] hooks;
            string hooksFolder;
            hooksFolder = Path.Combine(_configuration.ProgramRootFolder, HOOK_FOLDER);
            if ((!String.IsNullOrEmpty(_configuration.ProgramRootFolder)) && (Directory.Exists(hooksFolder)))
            {
                hooks = GetExistingHooks(hooksFolder);
                if (!hooks.Any(x => x.Value))
                {
                    MessageBox.Show("Currently, there are no Git hooks set up.", "Open Git Hooks");
                }
                else
                {
                    foreach (KeyValuePair<string, bool> hook in hooks)
                    {
                        if (hook.Value)
                        {
                            EditFileInNotepad(Path.Combine(hooksFolder, hook.Key), false);
                        }
                    }
                }
            } else
            {
                MessageBox.Show("Please configure a Program Root Folder first.", "Open Git Hooks");
            }
        }

        /// <summary>
        /// Edits a file in notepad.Returns when the user exits notepad.
        /// </summary>
        /// <param name="file"></param>
        /// <param name="waitForExit"></param>
        private void EditFileInNotepad(string file, bool waitForExit)
        {
            System.Diagnostics.Process myProcess;
            System.Diagnostics.ProcessStartInfo myProcessStartInfo;
            string sProgram;

            try
            {
                sProgram = "Notepad.exe";
                myProcess = new System.Diagnostics.Process();
                myProcessStartInfo = new System.Diagnostics.ProcessStartInfo(sProgram);
                myProcess.StartInfo = myProcessStartInfo;
                myProcessStartInfo.Arguments = file;
                myProcess.Start();
                if (waitForExit)
                {
                    myProcess.WaitForExit();
                }
            }
            catch (System.Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
            return;
        }

        private void mnuDeleteGitHooks_Click(object sender, EventArgs e)
        {
            KeyValuePair<string, bool>[] hooks;
            string hookFile;
            string hooksFolder;
            hooksFolder = Path.Combine(_configuration.ProgramRootFolder, HOOK_FOLDER);
            if ((!String.IsNullOrEmpty(_configuration.ProgramRootFolder)) && (Directory.Exists(hooksFolder)))
            {
                hooks = GetExistingHooks(hooksFolder);
                if (!hooks.Any(x=>x.Value))
                {
                    MessageBox.Show("No git hooks are currently set up.", "Delete Git Hooks");
                } else
                {
                    if (MessageBox.Show("Are you sure you want to delete the git hook files?", "Delete Git Hook Files", MessageBoxButtons.YesNo) == DialogResult.Yes)
                    {
                        try
                        {

                            foreach (KeyValuePair<string, bool> hook in hooks)
                            {
                                if (hook.Value)
                                {
                                    hookFile = Path.Combine(hooksFolder, hook.Key);
                                    File.Delete(hookFile);
                                }
                            }
                        }
                        catch (Exception)
                        {

                        }
                    }                    
                }                
            }
            else
            {
                MessageBox.Show("Please configure a Program Root Folder first.", "Delete Git Hooks");
            }
        }
    }
}
