﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace VisualGitSafe
{
    
    public partial class frmConfiguration : Form
    {
        private Configuration _configuration;
        private bool _allowClose;
        private AspectRatio _aspectRatio;
        public frmConfiguration()
        {
            _configuration = new Configuration();
            _allowClose = false;
            InitializeComponent();
            LoadAspectRatio();
        }
        
        public frmConfiguration(Configuration configuration)
        {
            _configuration = configuration;
            if (_configuration == null)
            {
                _configuration = new Configuration();
            }
            _allowClose = false;
            InitializeComponent();
            txtYourName.Text = _configuration.YourName;
            txtProgramRootFolder.Text = _configuration.ProgramRootFolder;
            txtSharedFolder.Text = _configuration.SharedFolder;
            chkMakeEditedFilesReadOnly.Checked = _configuration.MakeEditedFilesReadOnly;
            txtFileExtensions.Text = _configuration.Extensions;
            LoadAspectRatio();
        }

        private void LoadAspectRatio()
        {
            _aspectRatio = new AspectRatio(this);
            _aspectRatio.AddControl(grpProgramRootFolder, 0, 1, 0, 0);
            _aspectRatio.AddControl(txtProgramRootFolder, 0, 1, 0, 0);
            _aspectRatio.AddControl(btnBrowseRootFolder, 1, 0, 0, 0);
            _aspectRatio.AddControl(grpYourName, 0, 1, 0, 0);
            _aspectRatio.AddControl(txtYourName, 0, 1, 0, 0);
            _aspectRatio.AddControl(grpSharedFolder, 0, 1, 0, 0);
            _aspectRatio.AddControl(txtSharedFolder, 0, 1, 0, 0);
            _aspectRatio.AddControl(btnBrowseSharedFolder, 1, 0, 0, 0);
            _aspectRatio.AddControl(grpFileExtensions, 0, 1, 0, 0);
            _aspectRatio.AddControl(txtFileExtensions, 0, 1, 0, 0);
            _aspectRatio.AddControl(btnExtensionDefault, 1, 0, 0, 0);
            _aspectRatio.AddControl(grpOptions, 0, 1, 0, 0);
        }

        private void btnBrowseRootFolder_Click(object sender, EventArgs e)
        {
            Browse(grpProgramRootFolder);
        }

        private void Browse(GroupBox group)
        {
            TextBox textBox = GetTextBox(group);
            FolderBrowserDialog dialog = new FolderBrowserDialog();
            DialogResult result;
            dialog.SelectedPath = textBox.Text;
            dialog.Description = $"Browse {group.Text}";
            result = dialog.ShowDialog();
            if (result == DialogResult.OK)
            {
                textBox.Text = dialog.SelectedPath;
            }
        }

        private TextBox GetTextBox(GroupBox group)
        {
            foreach(Control loop in group.Controls)
            {
                if (loop.GetType().ToString().Contains("TextBox"))
                {
                    return (TextBox)loop;
                }
            }
            return null;
        }

        private void btnBrowseSharedFolder_Click(object sender, EventArgs e)
        {
            Browse(grpSharedFolder);
        }

        private void SaveField(ref string target, string source)
        {
            if (target != source)
            {
                target = source;
                _configuration.IsDirty = true;
            }
        }

        private void SaveField(ref bool target, bool source)
        {
            if (target != source)
            {
                target = source;
                _configuration.IsDirty = true;
            }
        }

        private void frmConfiguration_FormClosing(object sender, FormClosingEventArgs e)
        {
            SaveField(ref _configuration.YourName, txtYourName.Text);
            SaveField(ref _configuration.ProgramRootFolder, txtProgramRootFolder.Text);
            SaveField(ref _configuration.SharedFolder, txtSharedFolder.Text);
            SaveField(ref _configuration.MakeEditedFilesReadOnly, chkMakeEditedFilesReadOnly.Checked);
            SaveField(ref _configuration.Extensions, txtFileExtensions.Text);
        }

        private void btnExtensionDefault_Click(object sender, EventArgs e)
        {
            txtFileExtensions.Text = Configuration.DefaultExtensions;
        }
    }
}
